import re
from typing import Dict, List, Optional, Tuple

from bs4 import BeautifulSoup, NavigableString, Tag


def find_string_after_tag(t: Tag, text: str):
    return (
        isinstance(t.next_element, NavigableString)
        and t.next_element is not None
        and re.search(text, t.next_element)
    )


def chapters(t: Tag):
    return find_string_after_tag(t, r"CHAPTER")


def articles(t: Tag):
    return find_string_after_tag(t, r"ARTICLE")


def make_container_item(t: Tag):
    target = t.next_element.next_element
    if isinstance(target, Tag) and target.name == "br":
        return {
            "container id": t["id"],
            "component": str(t.next_element),
            "label": str(target.next_element),
        }
    else:
        return {
            "container_id": t["id"],
            "component": t.text.title(),
            "label": target.text.title()
            if isinstance(target, Tag)
            else target.title(),
        }


def get_containers(
    html: BeautifulSoup, first_section: str
) -> Optional[List[Tag]]:
    # slice with start of body to the start of the first section
    # used for checking if containers exist (e.g. interim slice)
    # if there are no containers between the interim slice, there should be no containers;
    # this prevents issues like the TRAIN LAW which amends CHAPTERS / ARTICLES
    interim_container = str(html)
    end_index = interim_container.index(first_section)
    cut_container = interim_container[:end_index]
    interim = BeautifulSoup(cut_container, "html5lib")
    if not interim(chapters) and not interim(articles):
        return None  # no containers found
    return html(chapters) or html(articles)  # container found


def fetch_containers(
    html: BeautifulSoup, first_section: str
) -> Tuple[BeautifulSoup, Optional[List[Tag]], Optional[List[Dict[str, str]]]]:
    if not (containers := get_containers(html, first_section)):
        return html, containers, None
    else:
        for index, parent_tag in enumerate(containers):
            parent_tag["id"] = f"{index}"
            parent_tag["data-type"] = "container"
            if (
                parent_tag.next_sibling
                and parent_tag.next_sibling.name == "br"
            ):  # deal with improper formats
                parent_tag.next_sibling.decompose()
        container_list = [make_container_item(i) for i in containers]
        return html, containers, container_list
