import json
import re
from typing import Optional, Tuple

from bs4 import BeautifulSoup, NavigableString, Tag

from .containers import fetch_containers
from .units import fetch_units


def find_string_after_tag(t: Tag, text: str):
    return (
        isinstance(t.next_element, NavigableString)
        and t.next_element is not None
        and re.search(text, t.next_element)
    )


def chapters(t: Tag):
    return find_string_after_tag(t, r"CHAPTER")


def articles(t: Tag):
    return find_string_after_tag(t, r"ARTICLE")


def full_section(t: Tag):
    return find_string_after_tag(t, r"^\s*SECTION")


def has_enacting(t: Tag):
    return find_string_after_tag(t, r"^Be\s+it\s+enacted")


def approval_mark(t: Tag):
    return find_string_after_tag(t, r"^Approved\,")


def modify_first_section(
    html: BeautifulSoup,
) -> Tuple[BeautifulSoup, Optional[str]]:
    if not (first_section := html.find(full_section)):
        return html, None
    first_section["id"] = "1"
    first_section["data-type"] = "unit"
    return html, str(first_section)


def remove_footer(text: str) -> str:
    soup = BeautifulSoup(text, "html5lib")
    for ender in soup("hr"):  # see cases like 86448
        [i.decompose() for i in ender.next_siblings if isinstance(i, Tag)]
        ender.decompose()
    return str(soup)


def get_signers_start(text: str) -> Tuple[str, Optional[str]]:
    soup = BeautifulSoup(text, "html5lib")
    if not (marker := soup.find(approval_mark)):
        return text, None  # text not modified
    new_tag = soup.new_tag("p")
    new_tag["data-type"] = "terminal"
    marker.wrap(new_tag)  # wrap approval string
    return str(soup), str(new_tag)  # return modified text with string


def separator(content: str) -> Tuple[str, Optional[str]]:
    content, signers_marker = get_signers_start(content)
    if not signers_marker:
        return content, None
    i = content.index(signers_marker)
    return content[:i], content[i:]


def meta_ra(raw: str):
    # get initial parameters
    provisions, signatories = separator(remove_footer(raw))
    html, first = modify_first_section(BeautifulSoup(provisions, "html5lib"))
    enacter = e.text if (e := html.find(has_enacting)) else None

    # return incomplete data
    if not first:
        return {
            "enacting_clause": enacter,
            "signatory_content": signatories,
            "first_section": None,
            "containers": None,
            "units": None,
            "body_statute": str(html),
        }

    # return complete data
    soup, interim_containers, container_list = fetch_containers(html, first)
    soup, unit_list = fetch_units(soup, interim_containers)
    return {
        "enacting_clause": enacter,
        "signatory_content": signatories,
        "first_section": first,
        "containers": json.dumps(container_list),
        "units": json.dumps(unit_list),
        "body_statute": str(soup),
    }
